#![allow(clippy::option_if_let_else)]

use convert_case::{Case, Casing};
use darling::{
    ast::{self, Data},
    util::{Override, PathList},
    Error, FromDeriveInput, FromField, FromMeta,
};
use proc_macro2::{Span, TokenStream};
use quote::{format_ident, quote};
use syn::parse_quote;

#[derive(Debug, FromField)]
#[darling(attributes(config))]
#[darling(forward_attrs(doc))]
pub struct ConfigField {
    ident: Option<syn::Ident>,
    vis: syn::Visibility,
    ty: syn::Type,
    attrs: Vec<syn::Attribute>,
    min: Option<Override<syn::Expr>>,
    max: Option<Override<syn::Expr>>,
    default: Option<Override<syn::Expr>>,
    #[darling(skip)]
    description: Option<syn::Expr>,
    #[darling(skip)]
    parent_ident: Option<syn::Ident>,
}

impl ConfigField {
    pub(crate) fn needs_newtype(&self) -> bool {
        self.helper_keys().len() > 1
    }

    pub(crate) fn helper(&self, attrs: &[syn::Attribute]) -> (Option<TokenStream>, syn::Type) {
        if self.needs_newtype() {
            let derives = attrs
                .iter()
                .find(|attr| attr.path().is_ident("derive"))
                .map(|attr| {
                    let mut derives = (*PathList::from_meta(&attr.meta).unwrap()).clone();
                    derives.retain(|derive| {
                        derive.segments.last().map(|s| {
                            let derive = s.ident.to_string();
                            derive.contains("Tree")
                                || derive.contains("Serialize")
                                || derive.contains("Deserialize")
                        }) == Some(false)
                    });
                    quote! {#[derive(#(#derives,)*)]}
                });
            let new_type_ident = self.helper_ident();
            let vis = &self.vis;
            let ty = &self.ty;

            let new = self.helper_new();
            let default = self.helper_default();
            let serde = self.helper_serde();
            let tree = self.helper_tree();

            (
                Some(quote! {
                    #derives
                    #vis struct #new_type_ident(#ty);

                    #new
                    #default
                    #serde
                    #tree
                }),
                parse_quote!(#new_type_ident),
            )
        } else {
            (None, self.ty.clone())
        }
    }

    pub(crate) fn helper_ident(&self) -> syn::Ident {
        format_ident!(
            "__{}{}",
            self.parent_ident
                .as_ref()
                .unwrap_or(&syn::Ident::new("Struct", Span::mixed_site())),
            self.ident
                .as_ref()
                .map_or("Field".to_owned(), |v| v.to_string().to_case(Case::Pascal))
        )
    }

    pub(crate) fn has_custom_limits(&self) -> bool {
        !((self.min.is_none() || self.min == Some(Override::Inherit))
            && (self.max.is_none() || self.max == Some(Override::Inherit)))
    }

    pub(crate) fn helper_new(&self) -> TokenStream {
        let ident = self.helper_ident();
        let ty = &self.ty;
        if self.has_custom_limits() {
            let min = self.min.as_ref().and_then(|v| v.as_ref().explicit());
            let max = self.max.as_ref().and_then(|v| v.as_ref().explicit());
            quote! {
                impl #ident {
                    pub fn new(value: #ty) -> Option<Self> {
                        if (#min..=#max).contains(&value) {
                            Some(Self(value))
                        } else {
                            None
                        }
                    }

                    #[allow(dead_code)]
                    pub const unsafe fn new_unchecked(value: #ty) -> Self {
                        Self(value)
                    }
                }

                impl ::core::ops::Deref for #ident {
                    type Target = #ty;
                    fn deref(&self) -> &Self::Target {
                        &self.0
                    }
                }
            }
        } else {
            quote! {
                impl #ident {
                    pub const fn new(value: #ty) -> Self {
                        Self(value)
                    }
                }

                impl ::core::ops::Deref for #ident {
                    type Target = #ty;
                    fn deref(&self) -> &Self::Target {
                        &self.0
                    }
                }

                impl ::core::ops::DerefMut for #ident {
                    fn deref_mut(&mut self) -> &mut Self::Target {
                        &mut self.0
                    }
                }
            }
        }
    }

    pub(crate) fn helper_default(&self) -> Option<TokenStream> {
        self.default.as_ref().map(|default| {
            let ident = self.helper_ident();
            let default_default = parse_quote!(default::Default());
            let default = default.as_ref().unwrap_or(&default_default);
            quote! {
                impl ::core::default::Default for #ident {
                    fn default() -> Self {
                        Self(#default)
                    }
                }
            }
        })
    }

    pub(crate) fn helper_serde(&self) -> TokenStream {
        let ident = self.helper_ident();
        let conversion = if self.has_custom_limits() {
            quote! {
                Self::new(value).ok_or_else(|| {
                    <D::Error as ::serde::de::Error>::custom("checking value bounds")
                })
            }
        } else {
            quote! {
                Ok(Self::new(value))
            }
        };
        let ty = &self.ty;
        quote! {
            impl ::serde::Serialize for #ident {
                fn serialize<S>(&self, serializer: S) -> ::core::result::Result::<S::Ok, S::Error>
                where
                    S: ::serde::Serializer,
                {
                    self.0.serialize(serializer)
                }
            }

            impl<'de> ::serde::Deserialize<'de> for #ident {
                fn deserialize<D>(deserializer: D) -> ::core::result::Result<Self, D::Error>
                where
                    D: ::serde::Deserializer<'de>,
                {
                    type T = #ty;
                    let value = T::deserialize(deserializer)?;
                    #conversion
                }
            }
        }
    }

    pub(crate) fn helper_tree(&self) -> TokenStream {
        let mut tokens = self.helper_tree_key();
        tokens.extend(self.helper_tree_serialize());
        tokens.extend(self.helper_tree_deserialize());
        tokens.extend(self.helper_tree_any());
        tokens
    }

    pub(crate) fn helper_keys(&self) -> Vec<(&'static str, syn::Expr)> {
        macro_rules! field_to_key {
            ($field: ident, $default: expr) => {
                self.$field.as_ref().map(|value| {
                    let ty = &self.ty;
                    (stringify!($field),
                    value.clone().explicit().map(|val| {
                        parse_quote!({
                            fn type_ascribe(val: #ty) -> #ty {val}
                            type_ascribe(#val)
                        })
                    }).unwrap_or_else(|| {
                        parse_quote!(#ty::$default)
                    }))
                })
            };
            ($field: ident) => {
                self.$field.as_ref().map(|value| {
                    (stringify!($field), value.clone())
                })
            }
        }
        [
            Some(("value", parse_quote!(self))),
            field_to_key!(default, default()),
            field_to_key!(min, MIN),
            field_to_key!(max, MAX),
            field_to_key!(description),
        ]
        .into_iter()
        .flatten()
        .collect::<Vec<_>>()
    }

    pub(crate) fn helper_tree_key(&self) -> TokenStream {
        let ident = self.helper_ident();
        let keys = self.helper_keys();
        let num_keys = keys.len();
        let keys = keys.iter().map(|(name, _)| name);
        let max_length = keys.clone().map(|v| v.len()).max();
        quote! {
            impl ::miniconf::KeyLookup for #ident {
                const LEN: usize = #num_keys;
                const NAMES: &'static [&'static str] = &[#(#keys,)*];

                fn name_to_index(value: &str) -> Option<usize> {
                    Self::NAMES.iter().position(|name| *name == value)
                }
            }

            impl ::miniconf::TreeKey<1> for #ident {
                fn metadata() -> ::miniconf::Metadata {
                    let mut metadata = ::miniconf::Metadata::default();
                    metadata.max_depth = 1;
                    metadata.count = #num_keys;
                    metadata.max_length = #max_length;
                    metadata
                }

                fn traverse_by_key<K, F, E>(mut keys: K, mut func: F) -> ::core::result::Result<usize, ::miniconf::Error<E>>
                where
                    K: ::miniconf::Keys,
                    // Writing this to return an iterator instead of using a callback
                    // would have worse performance (O(n^2) instead of O(n) for matching)
                    F: FnMut(usize, Option<&'static str>, usize) -> ::core::result::Result<(), E>,
                {
                    let ::core::result::Result::Ok(key) = keys.next::<Self>() else { return ::core::result::Result::Ok(0) };
                    let index = ::miniconf::Key::find::<Self>(&key).ok_or(::miniconf::Traversal::NotFound(1))?;
                    let name = <Self as ::miniconf::KeyLookup>::NAMES
                        .get(index)
                        .ok_or(::miniconf::Traversal::NotFound(1))?;
                    func(index, Some(name), #num_keys).map_err(|err| ::miniconf::Error::Inner(1, err))?;
                    ::core::result::Result::Ok(1)
                }
            }
        }
    }

    pub(crate) fn helper_tree_serialize(&self) -> TokenStream {
        let matches = self
            .helper_keys()
            .iter()
            .enumerate()
            .map(|(i, (_, expr))| {
                quote! {
                    #i => ::serde::Serialize::serialize(&#expr, ser).map_err(|err| ::miniconf::Error::Inner(0, err)),
                }
            })
            .collect::<Vec<_>>();
        let ident = self.helper_ident();
        quote! {
            impl ::miniconf::TreeSerialize<1> for #ident {
                fn serialize_by_key<K, S>(
                    &self,
                    mut keys: K,
                    ser: S,
                ) -> ::core::result::Result<usize, ::miniconf::Error<S::Error>>
                where
                    K: ::miniconf::Keys,
                    S: ::serde::Serializer,
                {
                    let ::core::result::Result::Ok(key) = keys.next::<Self>() else {
                        return ::serde::Serialize::serialize(&self, ser).map_err(|err| ::miniconf::Error::Inner(1, err)).map(|_| 1);
                    };
                    let index = ::miniconf::Key::find::<Self>(&key).ok_or(::miniconf::Traversal::NotFound(1))?;
                    if !keys.finalize() {
                        return ::core::result::Result::Err(::miniconf::Traversal::TooLong(1).into());
                    }
                    match index {
                        #(#matches)*
                        _ => unreachable!(),
                    }?;
                    Ok(0)
                }
            }
        }
    }

    pub(crate) fn helper_tree_deserialize(&self) -> TokenStream {
        let ident = self.helper_ident();
        let num_keys = self.helper_keys().len();
        quote! {
            impl<'de> ::miniconf::TreeDeserialize<'de, 1> for #ident {
                fn deserialize_by_key<K, D>(
                    &mut self,
                    mut keys: K,
                    de: D,
                ) -> ::core::result::Result<usize, ::miniconf::Error<D::Error>>
                where
                    K: ::miniconf::Keys,
                    D: ::serde::Deserializer<'de>,
                {
                    let ::core::result::Result::Ok(key) = keys.next::<Self>() else {
                        *self = <Self as ::serde::Deserialize>::deserialize(de).map_err(|err| ::miniconf::Error::Inner(0, err))?;
                        return ::core::result::Result::Ok(0);
                    };
                    let index = ::miniconf::Key::find::<Self>(&key).ok_or(::miniconf::Traversal::NotFound(1))?;
                    if !keys.finalize() {
                        return ::core::result::Result::Err(::miniconf::Traversal::TooLong(1).into());
                    }
                    match index {
                        0 => {
                            *self = <Self as ::serde::Deserialize>::deserialize(de).map_err(|err| ::miniconf::Error::Inner(0, err))?;
                            Ok(0)
                        }
                        ,
                        1..=#num_keys => ::core::result::Result::Err(::miniconf::Traversal::Access(0, "Cannot write limits").into()),
                        _ => unreachable!(),
                    }
                }
            }
        }
    }

    pub(crate) fn helper_tree_any(&self) -> TokenStream {
        let ident = self.helper_ident();
        let num_keys = self.helper_keys().len();
        let ref_mut = if self.has_custom_limits() {
            quote! {::core::result::Result::Err(::miniconf::Traversal::Access(0, "field has custom limits"))}
        } else {
            quote!(::core::result::Result::Ok(&mut *self))
        };
        quote! {
            impl ::miniconf::TreeAny<1> for #ident {
                fn ref_any_by_key<K>(&self, mut keys: K) -> ::core::result::Result<&dyn ::core::any::Any, ::miniconf::Traversal>
                where
                    K: ::miniconf::Keys,
                {
                    let ::core::result::Result::Ok(key) = keys.next::<Self>() else {
                        return ::core::result::Result::Ok(&*self);
                    };
                    let index = ::miniconf::Key::find::<Self>(&key).ok_or(::miniconf::Traversal::NotFound(1))?;
                    if !keys.finalize() {
                        return ::core::result::Result::Err(::miniconf::Traversal::TooLong(1));
                    }
                    match index {
                        0 => ::core::result::Result::Ok(&*self),
                        1..#num_keys => ::core::result::Result::Err(::miniconf::Traversal::Access(1, "cannot return reference to local variable")),
                        _ => unreachable!(),
                    }
                }

                fn mut_any_by_key<K>(&mut self, mut keys: K) -> ::core::result::Result<&mut dyn ::core::any::Any, ::miniconf::Traversal>
                where
                    K: ::miniconf::Keys,
                {
                    let ::core::result::Result::Ok(key) = keys.next::<Self>() else {
                        return #ref_mut;
                    };
                    let index = ::miniconf::Key::find::<Self>(&key).ok_or(::miniconf::Traversal::NotFound(1))?;
                    if !keys.finalize() {
                        return ::core::result::Result::Err(::miniconf::Traversal::TooLong(1));
                    }
                    match index {
                        0 => #ref_mut,
                        1..#num_keys => ::core::result::Result::Err(::miniconf::Traversal::Access(1, "cannot return reference to local variable")),
                        _ => unreachable!(),
                    }
                }
            }
        }
    }
}

#[derive(Debug, FromDeriveInput)]
#[darling(forward_attrs(derive))]
#[darling(supports(struct_any))]
pub struct Config {
    ident: syn::Ident,
    // generics: syn::Generics,
    // vis: syn::Visibility,
    data: ast::Data<(), ConfigField>,
    pub attrs: Vec<syn::Attribute>,
}

impl Config {
    pub(crate) fn parse(input: &syn::DeriveInput) -> Result<Self, Error> {
        let mut config = Self::from_derive_input(input)?;
        let ident = config.ident.clone();
        let fields = config.fields_mut();
        for field in fields.iter_mut() {
            field.description = field
                .attrs
                .iter()
                .find(|attr| attr.path().is_ident("doc"))
                .map(|attr| {
                    let description = syn::LitStr::from_meta(&attr.meta).unwrap().value();
                    let description = description.trim();
                    parse_quote!(#description)
                });
            field.parent_ident = Some(ident.clone());
        }
        Ok(config)
    }

    pub(crate) fn fields(&self) -> &Vec<ConfigField> {
        let Data::Struct(fields) = &self.data else {
            unreachable!()
        };
        &fields.fields
    }

    pub(crate) fn fields_mut(&mut self) -> &mut Vec<ConfigField> {
        let Data::Struct(fields) = &mut self.data else {
            unreachable!()
        };
        &mut fields.fields
    }
}
